import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from os import path
from linfa.mlp import FNN

torch.set_default_tensor_type(torch.DoubleTensor)

class Discrepancy(object):
    """Class to create surrogate models for discrepancy estimatio
       TO BE COMPLETED!!!
    """
    def __init__(self, model_name,
                       lf_model,
                       input_size,
                       output_size,
                       var_grid_in,
                       var_grid_out,
                       dnn_arch=None, 
                       dnn_activation='relu', 
                       model_folder='./', 
                       surrogate=None, 
                       device='cpu'):
        
        self.device = device
        self.input_size = input_size
        self.output_size = output_size
        self.model_name = model_name
        self.model_folder = model_folder
        self.is_trained = False

        # Assign LF model
        self.lf_model = lf_model

        # Store variable grid locations
        self.var_grid_in=var_grid_in
        # Output variables - multiple noisy observations 
        # are available for each combination of variables
        self.var_grid_out=var_grid_out

        # Input/output statistics
        self.var_in_avg = torch.mean(var_grid_in,dim=0)
        if(len(self.var_grid_in) == 1):
            self.var_in_std = torch.zeros_like(self.var_in_avg)
        else:
            self.var_in_std = torch.std(var_grid_in,dim=0)
        # If there are multiple outputs, we will define one file for each output
        self.var_out_avg = torch.mean(var_grid_out)
        self.var_out_std = torch.std(var_grid_out)

        # Create surrogate
        self.surrogate = FNN(input_size, output_size, arch=dnn_arch, device=self.device, init_zero=True) if surrogate is None else surrogate
        
    def surrogate_save(self):
        """Save surrogate model to [self.name].sur and [self.name].npz
        
        Returns:
            None

        """
        # Save model state dictionary
        torch.save(self.surrogate.state_dict(), self.model_folder + self.model_name + '.sur')

    def surrogate_load(self):
        """Load surrogate model from [self.name].sur and [self.name].npz
        
        Returns:
            None
        """
        # Read back the state dictionary from file
        self.surrogate.load_state_dict(torch.load(self.model_folder + self.model_name + '.sur'))

    def update(self, batch_x, max_iters=10000, lr=0.01, lr_exp=0.999, record_interval=50, store=True, reg=False, reg_penalty=0.0001):
        """Train surrogate model with pre-grid.

        """
        print('')
        print('--- Training model discrepancy')
        print('')

        self.is_trained = True

        # LF model output at the current batch
        y = self.lf_model(batch_x) 

        # Standardize inputs and outputs
        if(len(self.var_grid_in) == 1):
            var_grid = torch.zeros_like(self.var_grid_in)
        else:
            var_grid = (self.var_grid_in - self.var_in_avg) / self.var_in_std
        # The output is the discrepancy
        var_out = self.var_grid_out - torch.mean(y,dim=1).unsqueeze(1)
        # Update output stats
        # self.var_out_avg = torch.mean(var_out)
        # self.var_out_std = torch.std(var_out)
        # Rescale outputs
        # var_out = (var_out - self.var_out_avg) / self.var_out_std

        # Set optimizer and scheduler
        optimizer = torch.optim.RMSprop(self.surrogate.parameters(), lr=lr)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, lr_exp)

        for i in range(max_iters):
            # Set surrogate in training mode
            self.surrogate.train()          
            # Surrogate returns a table with rows as batches and columns as variables considered            
            disc = self.surrogate(var_grid)

            # Compute loss averaged over batches/variables
            # Mean over the columns (batches)
            # Mean over the rows (variables)
            # Also we need to account for the number of repeated observations
            loss = torch.tensor(0.0)            
            # Loop over the number of observations
            for loopA in range(var_out.size(1)):
                loss += torch.sum((disc.flatten() - var_out[:,loopA]) ** 2)
                if reg:
                    reg_loss = 0
                    for param in self.surrogate.parameters():
                        reg_loss += torch.abs(param).sum() * reg_penalty
                    loss += reg_loss
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            scheduler.step()

            if i % record_interval == 0:
                if reg:
                    print('DISC: it: %7d | loss: %8.3e | reg_loss: %8.3e' % (i, loss, reg_loss))
                else:
                    print('DISC: it: %7d | loss: %8.3e' % (i, loss))
        print('')
        print('--- Surrogate model pre-train complete')
        print('')        
        if store:
            self.surrogate_save()

    def forward(self, var):
        """Function to evaluate the surrogate
        
        Args:
            x (torch.Tensor): Contains a matrix of model inputs in the form [data_num, feature_dim]

        Returns:
            Value of the surrogate at x.

        """
        if(len(self.var_grid_in) == 1):
            res = self.surrogate(torch.zeros_like(var))
        else:
            res = self.surrogate((var - self.var_in_avg) / self.var_in_std)

        if not(self.is_trained):
            zero_res = torch.zeros_like(res)
            return zero_res
        else:
            return res

def test_surrogate():
    
    import matplotlib.pyplot as plt
    from linfa.models.discrepancy_models import PhysChem

    # Set variable grid
    var_grid = [[350.0, 400.0, 450.0],
                [1.0, 2.0, 3.0, 4.0, 5.0]]

    # Create Model
    model = PhysChem(var_grid)

    # Generate true data
    model.genDataFile(dataFileNamePrefix='observations', use_true_model=True, store=True, num_observations=10)  

    # Get data from true model at the same TP conditions    
    var_data = np.loadtxt('observations.csv',skiprows=1,delimiter=',')
    var_data_in = torch.tensor(var_data[:,:2])
    var_data_out = torch.tensor(var_data[:,2:])
    
    # Define emulator and pre-train on global grid
    discrepancy = Discrepancy(model_name='discrepancy_test',
                              lf_model=model.solve_lf,
                              input_size=2,
                              output_size=1,
                              var_grid_in=var_data_in,
                              var_grid_out=var_data_out)
    
    # Create a batch of samples for the calibration parameters
    batch_x = model.defParams

    # Update the discrepancy model
    discrepancy.update(batch_x, max_iters=1000, lr=0.001, lr_exp=0.9999, record_interval=100)

    # Plot discrepancy
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.scatter(model.var_in[:,0].detach().numpy(), model.var_in[:,1].detach().numpy(), discrepancy.forward(model.var_in).detach().numpy(), marker='o')
    ax.set_xlabel('Temperature')
    ax.set_ylabel('Pressure')
    ax.set_zlabel('Coverage')
    plt.show()

# TEST SURROGATE
if __name__ == '__main__':
  
  test_surrogate()
    


import numpy as np
import torch

torch.set_default_tensor_type(torch.DoubleTensor)

def solve_ivp_s(func, t0, t_bound, y0, max_step, t_eval, batch_size, aux_size, device='cpu'):
    # PyTorch - Reimplementing scipy.integrate.solve_ivp (Runge-Kuta Method solving ODE) - Output compared and tested.
    n = int((t_bound - t0) / max_step)
    t = t0
    y = y0.double()
    y_rec = torch.zeros(len(t_eval), batch_size).to(device)
    aux_rec = torch.zeros(len(t_eval), aux_size, batch_size).to(device)
    t_rec = torch.zeros(len(t_eval)).to(device)
    i = 0
    for _ in range(n):
        res, aux = func(t, y)
        k1 = max_step * res
        res, aux = func(t + 0.5 * max_step, y + 0.5 * k1)
        k2 = max_step * res
        res, aux = func(t + 0.5 * max_step, y + 0.5 * k2)
        k3 = max_step * res
        res, aux = func(t + 0.5 * max_step, y + k3)  # I think this should be 1.0 and not 0.5!!!
        k4 = max_step * res
        delta = 1.0 / 6.0 * (k1 + 2 * k2 + 2 * k3 + k4)

        if t >= t_eval[i]:
            y_rec[i, :] = y
            aux_rec[i, :, :] = aux
            t_rec[i] = t
            i = i + 1
        y = y + delta
        t = t + max_step

    y_rec[len(t_eval) - 1, :] = y
    aux_rec[len(t_eval) - 1, :, :] = aux
    t_rec[len(t_eval) - 1] = t
    return t_rec, y_rec, aux_rec


def trapz(t, y):
    # Pytorch - Numerical Integration: Trapezoidal Rule
    S = list(y.size())
    t0 = t[:S[0] - 1]
    t1 = t[1:]
    y0 = y[:S[0] - 1, :]
    y1 = y[1:, :]
    return torch.sum((y1 + y0) / 2.0 * (t1 - t0).reshape(S[0] - 1, 1), 0)


class circuitModel():

    def __init__(self, numParam, numState, numAuxState, numOutputs,
                 parName, limits, defParam,
                 cycleTime, totalCycles, forcing=None, device='cpu'):
        
        # Assign device
        self.device = device
        # Time integration parameters
        self.cycleTime = cycleTime
        self.totalCycles = totalCycles
        # Forcing
        self.forcing = forcing
        # Init parameters
        self.numParam = numParam
        self.numState = numState
        self.numAuxState = numAuxState
        self.numOutputs = numOutputs
        self.parName = parName

        self.stdRatio = 0.05
        self.limits = limits
        self.mmHgToBarye = 1333.22
        self.defParam = defParam
        self.defOut = self.solve_t(self.defParam, y0=None)
        self.data = None

    def evalDeriv_t(self, t, y, params):
        # PyTorch - Computing Derivative for the model. See specific definition in subclasses of RC and RCR
        pass

    def postProcess_t(self, t, y, aux, start, stop):
        # PyTorch - Computing (min, max, ave) output given solution of ODE. See specific definition in subclasses of RC and RCR
        pass

    def genDataFile(self, dataSize, dataFileName):
        # Scipy - Generate Data file: Given the solution def_out of default parameters,
        # sample data with mean def_out and cov matrix with diagonal std * def_out
        data = np.zeros((self.numOutputs, dataSize))
        # Get Standard Deviaitons using ratios
        stds = self.defOut * self.stdRatio
        for loopA in range(dataSize):
            # Get Default Paramters
            data[:, loopA] = self.defOut[0] + torch.randn(len(stds[0])) * stds
        self.data = data
        np.savetxt(dataFileName, data)

    def solve_t(self, params, y0=None):
        # Pytorch - Reimplementing solve: Support Multiple parameters
        batch_size = list(params.size())[0]
        if y0 is None: y0 = 55.0 * self.mmHgToBarye * torch.ones(batch_size).to(self.device)
        t_bound = self.totalCycles * self.cycleTime
        saveSteps = np.linspace(0.0, t_bound, 201, endpoint=True)
        odeSol_t, odeSol_y, odeSol_aux = solve_ivp_s(lambda t, y: self.evalDeriv_t(t, y, params.double()), 0.0, t_bound,
                                                     y0 * torch.ones(batch_size).to(self.device),
                                                     max_step=self.cycleTime / 1000.0, t_eval=saveSteps,
                                                     batch_size=batch_size, aux_size=self.numAuxState, device=self.device)
        start = len(saveSteps) - (len(saveSteps[saveSteps > (self.totalCycles - 1) * self.cycleTime]) + 1)
        stop  = len(saveSteps)

        return self.postProcess_t(odeSol_t, odeSol_y, odeSol_aux, start, stop)

class rcModel(circuitModel):

    def __init__(self, cycleTime, totalCycles, forcing=None, device='cpu'):
        # Init parameters
        numParam = 2
        numState = 1
        numAuxState = 4
        numOutputs = 3
        parName = ["R", "C"]
        limits = torch.Tensor([[100.0, 1500.0], [1.0e-5, 1.0e-2]]).to(device)
        defParam = torch.Tensor([[1000.0, 0.00005]]).to(device)
        #  Invoke Superclass Constructor
        super().__init__(numParam, numState, numAuxState, numOutputs,
                         parName, limits, defParam,
                         cycleTime, totalCycles, forcing, device=device)

    def evalDeriv_t(self, t, y, params):
        # Pytorch - Evaluate Derivative.
        R = params[:, 0]
        C = params[:, 1]
        Pd = torch.tensor(55 * self.mmHgToBarye).to(self.device)
        P1 = y

        # Interpolate forcing
        Q1 = np.interp(t % self.cycleTime, self.forcing[:, 0], self.forcing[:, 1])
        Q2 = (P1 - Pd) / R
        dP1dt = (Q1 - Q2) / C

        aux = torch.zeros(tuple((self.numAuxState,)) + tuple(dP1dt.size()))
        aux[0] = Pd
        aux[1] = Q1
        aux[2] = Q2
        return dP1dt, aux

    def postProcess_t(self, t, y, aux, start, stop):
        # PyTorch - Computing (Min, Max, Ave) tuple given solution of ODE
        return torch.stack([torch.min(y[start:stop, :], 0)[0] / self.mmHgToBarye,
                            torch.max(y[start:stop, :], 0)[0] / self.mmHgToBarye,
                            trapz(t[start:stop], y[start:stop, :]) / float(self.cycleTime) / self.mmHgToBarye],
                           dim=1)

class rcrModel(circuitModel):

    def __init__(self, cycleTime, totalCycles, forcing=None, device='cpu'):
        # Init parameters
        numParam = 3
        numState = 1
        numAuxState = 4
        numOutputs = 3
        parName = ["R1", "R2", "C"]
        limits = torch.Tensor([[100.0, 1500.0],
                               [100.0, 1500.0],
                               [1.0e-5, 1.0e-2]]).to(device)
        defParam = torch.Tensor([[1000.0, 1000.0, 0.00005]]).to(device)
        #  Invoke Superclass Constructor
        super().__init__(numParam, numState, numAuxState, numOutputs,
                         parName, limits, defParam,
                         cycleTime, totalCycles, forcing, device=device)
        self.stdRatio = 0.05

    def evalDeriv_t(self, t, y, params):
        R1 = params[:, 0]
        R2 = params[:, 1]
        C = params[:, 2]
        Pd = 55 * self.mmHgToBarye * torch.ones(params.shape[0]).to(self.device)
        P1 = y

        # Interpolate forcing
        Q1 = np.interp(t % self.cycleTime, self.forcing[:, 0], self.forcing[:, 1])
        P0 = P1 + R1 * Q1
        Q2 = (P1 - Pd) / R2
        dP1dt = (Q1 - Q2) / C

        aux = torch.zeros(tuple((self.numAuxState,)) + tuple(dP1dt.size()))
        aux[0] = Pd
        aux[1] = P0
        aux[2] = Q1
        aux[3] = Q2
        return dP1dt, aux

    def postProcess_t(self, t, y, aux, start, stop):
        return torch.stack([torch.min(aux[start:stop, 1, :], 0)[0] / self.mmHgToBarye,
                            torch.max(aux[start:stop, 1, :], 0)[0] / self.mmHgToBarye,
                            trapz(t[start:stop], aux[start:stop, 1, :]) / float(self.cycleTime) / self.mmHgToBarye],
                           dim=1)

# GEN DATA
if __name__ == '__main__':
  
  # model = rcModel()
  model = rcrModel()
  model.genDataFile(50, "../../resource/data/data_rcr.txt")


import sys, os
import numpy as np
import torch

torch.set_default_tensor_type(torch.DoubleTensor)

class Highdim:

    def __init__(self,device='cpu'):
        # Init parameters
        self.input_num = 5
        self.output_num = 4
        self.x0 = torch.Tensor([0.0838, 0.2290, 0.9133, 0.1524, 0.8258]).to(device)
        self.defParam = torch.Tensor([[2.75, -1.5, 0.25, -2.5, 1.75]]).to(device)
        self.RM = torch.Tensor([[1, 0, 0, 0], [1, 1, 0, 0], [0, 1, 1, 0], [0, 0, 1, 1], [0, 0, 0, 1]]).to(device) / np.sqrt(2.0)
        self.defOut = self.solve_t(self.defParam)
        self.stdRatio = 0.01
        self.data = None

    def genDataFile(self, dataSize=50, dataFileName="../../resource/data/data_highdim.txt", store=True):
        def_out = self.defOut[0]
        self.data = def_out + self.stdRatio * torch.abs(def_out) * torch.normal(0, 1, size=(dataSize, self.output_num))
        self.data = self.data.t().detach().numpy()
        if store: np.savetxt(dataFileName, self.data)
        return self.data

    def solve_t(self, params):
        return torch.matmul((2 * torch.abs(2 * self.x0 - 1) + torch.exp(params)) / (1 + torch.exp(params)), self.RM)

# GEN DATA
if __name__ == '__main__':
  
  model = Highdim()
  model.genDataFile(50)